/*
 * NVRAM variable manipulation (direct mapped flash)
 *
 * Copyright 2001-2004, Broadcom Corporation
 * All Rights Reserved.
 *
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 */

#include <typedefs.h>
#include <osl.h>
#include <bcmboard.h>
#include <bcmnvram.h>
#include <bcmendian.h>
#include <bcmutils.h>
#include <flash.h>
#include <flashutl.h>
#include <sbconfig.h>
#include <sbchipc.h>
#include <sbutils.h>

struct nvram_tuple * _nvram_realloc(struct nvram_tuple *t, const char *name, const char *value);
void _nvram_free(struct nvram_tuple *t);
int _nvram_read(void *buf);

extern char * _nvram_get(const char *name);
extern int _nvram_set(const char *name, const char *value);
extern int _nvram_unset(const char *name);
extern int _nvram_getall(char *buf, int count);
extern int _nvram_commit(struct nvram_header *header);
extern int _nvram_init(void);
extern void _nvram_exit(void);

static struct nvram_header *nvram_header = NULL;
static struct nvram_header *nvram_temp_header = NULL;
static 	ulong flash_base = 0;

#ifdef vxworks
#include <vxWorks.h>
#include <vxLib.h>
#include <taskLib.h>
static SEM_ID nvram_lock = 0;
#define NVRAM_LOCK()	do { if (nvram_lock) semTake(nvram_lock, WAIT_FOREVER); } while (0)
#define NVRAM_UNLOCK()	do { if (nvram_lock) semGive(nvram_lock); } while (0)
#else
#define NVRAM_LOCK()	do {} while (0)
#define NVRAM_UNLOCK()	do {} while (0)
#endif

/* Convenience */
#define KB * 1024
#define MB * 1024 * 1024

char *
nvram_get(const char *name)
{
	char *value;

	NVRAM_LOCK();
	value = _nvram_get(name);
	NVRAM_UNLOCK();

	return value;
}

int
nvram_getall(char *buf, int count)
{
	int ret;

	NVRAM_LOCK();
	ret = _nvram_getall(buf, count);
	NVRAM_UNLOCK();

	return ret;
}

int
nvram_set(const char *name, const char *value)
{
	int ret;

	NVRAM_LOCK();
	ret = _nvram_set(name, value);
	NVRAM_UNLOCK();

	return ret;
}

int
nvram_unset(const char *name)
{
	int ret;

	NVRAM_LOCK();
	ret = _nvram_unset(name);
	NVRAM_UNLOCK();

	return ret;
}

static bool
nvram_reset(void *sbh)
{
	chipcregs_t *cc;
	char *value;
	uint32 watchdog = 0, gpio;
	uint idx, msec;

	idx = sb_coreidx(sbh);

	/* Check if we were soft reset */
	if ((cc = sb_setcore(sbh, SB_CC, 0))) {
		watchdog = R_REG(&cc->intstatus) & 0x80000000;
		sb_setcoreidx(sbh, idx);
	}
	if (watchdog)
		return FALSE;

	value = nvram_get("reset_gpio");
	if (!value)
		return FALSE;

	gpio = (uint32) bcm_atoi(value);
	if (gpio > 7)
		return FALSE;

	/* Setup GPIO input */
	sb_gpioouten(sbh, (1 << gpio), 0);

	/* GPIO reset is asserted low */
	for (msec = 0; msec < 5000; msec++) {
		if (sb_gpioin(sbh) & (1 << gpio))
			return FALSE;
		OSL_DELAY(1000);
	}

	return TRUE;
}
	
#ifdef CFG_EMBEDDED_NVRAM
	extern unsigned char embedded_nvram[];
#endif
	
static struct nvram_header *  
find_nvram(bool embonly, bool *isemb, bool *iscopy)
{
	struct nvram_header *nvh;
	uint32 off, lim;

	*iscopy = FALSE;
	if (!embonly) {
		*isemb = FALSE;
		if (flash_base == CC_FLASH_BASE)
			lim = CC_FLASH_MAX;
		else
			lim = BCM_BOOTROM_MAX_SIZE;
		off = BCM_FLASH_MIN_SIZE;
		while (off <= lim) {
			/* Look for copy first */
			nvh = (struct nvram_header *) (flash_base + off - NVRAM_SEC_SIZE - NVRAM_SPACE);
			if (nvh->magic == NVRAM_MAGIC) {
				*iscopy = TRUE;
				return (nvh);
			}
			/* No copy found, look for the main */
			nvh = (struct nvram_header *) (flash_base + off - NVRAM_SPACE);
			if (nvh->magic == NVRAM_MAGIC)
				return (nvh);
			off <<= 1;
		};
	}

	/* Now check embedded nvram */
	*isemb = TRUE;
	nvh = (struct nvram_header *) (flash_base + (4 * 1024));
	if (nvh->magic == NVRAM_MAGIC)
		return (nvh);
	nvh = (struct nvram_header *) (flash_base + 1024);
	if (nvh->magic == NVRAM_MAGIC)
		return (nvh);
#ifdef _CFE_
#ifdef CFG_EMBEDDED_NVRAM
	nvh = (struct nvram_header *)embedded_nvram;
	if (nvh->magic == NVRAM_MAGIC)
		return (nvh);
#endif
#endif
	return (NULL);
}

int
nvram_init(void *sbh)
{
	int ret=0;
	uint idx;
	bool isemb;
	bool iscopy;

#ifdef vxworks
	/* Initialize private semaphore */
	if ((nvram_lock = semBCreate(SEM_Q_FIFO, SEM_FULL)) == NULL) {
		printf("nvram_init: semBCreate failed\n");
		return ERROR;
	}
#endif

	idx = sb_coreidx(sbh);
	if (sb_setcore(sbh, SB_CC, 0) != NULL) {
		flash_base = CC_FLASH_BASE;
		sb_setcoreidx(sbh, idx);
	} else
		flash_base = BCM_BOOTROM_BASE;

	/* Temporarily initialize with embedded NVRAM */
	nvram_header = find_nvram(TRUE, &isemb, &iscopy);
	ret = _nvram_init();
	if (ret == 0) {
		/* Restore defaults from embedded NVRAM if button held down */
		if (nvram_reset(sbh)) {
			return NV_INIT_RESET;
		}
		_nvram_exit();
	}

	/* Find NVRAM */
	nvram_header = find_nvram(FALSE, &isemb, &iscopy);
	ret = _nvram_init();
	if (ret == 0) {
		if (nvram_header && isemb) {
			/*
			We are using embedded NVRAM, check if the main region has
			a pattern, which tells us that this is the very first time
			the system is booting.
			*/
			struct nvram_header *nvh;
			uint32 off, lim;
			uint32 pattern;
			char *patternStr;
			
			/* Default return values is no NVRAM region found */
			ret = NV_INIT_NONV;

			/* check if goto_mfgstate_pattern is defined */
			if(strcmp((patternStr=nvram_safe_get("goto_mfgstate_pattern")), "")) {
				pattern = bcm_strtoul(patternStr, NULL, 0);

				if (flash_base == CC_FLASH_BASE)
					lim = CC_FLASH_MAX;
				else
					lim = BCM_BOOTROM_MAX_SIZE;
				off = BCM_FLASH_MIN_SIZE;
				while (off <= lim) {
					nvh = (struct nvram_header *) (flash_base + off - NVRAM_SPACE);
					if ((nvh->magic == pattern) && 
							(nvh->crc_ver_init == pattern)) {
						/*
						Run-time NVRAM region has pre-defined pattern, hence
						this is the first boot.
						*/
						ret = NV_INIT_MFGSTATE;
						break;
					}
					off <<= 1;
				};
			}
		} else if (nvram_header && iscopy) {
			/* Find out if main differs from copy. */
			nvram_temp_header = (struct nvram_header *) ((char *)nvram_header + NVRAM_SEC_SIZE);
			if ((nvram_temp_header != NULL) &&
					(nvram_temp_header->crc_ver_init != nvram_header->crc_ver_init)) {
				ret = NV_INIT_RESMAIN;	/* Restore copy to main. */
			}
		} else if (nvram_header) {
			/* Main NVRAM used, copy must be bad. */
			ret = NV_INIT_RESCOPY; /* Backup copy from main. */
		}
	}

	return ret;
}

void
nvram_exit(void)
{
	_nvram_exit();
}

int
_nvram_read(void *buf)
{
	uint32 *src, *dst;
	uint i;

	if (!nvram_header)
		return -19; /* -ENODEV */

	src = (uint32 *) nvram_header;
	dst = (uint32 *) buf;

	for (i = 0; i < sizeof(struct nvram_header); i += 4)
		*dst++ = *src++;

	for (; i < nvram_header->len && i < NVRAM_SPACE; i += 4)
		*dst++ = ltoh32(*src++);

	return 0;
}

struct nvram_tuple *
_nvram_realloc(struct nvram_tuple *t, const char *name, const char *value)
{
	if (!(t = MALLOC(sizeof(struct nvram_tuple) + strlen(name) + 1 + strlen(value) + 1)))
		return NULL;

	/* Copy name */
	t->name = (char *) &t[1];
	strcpy(t->name, name);

	/* Copy value */
	t->value = t->name + strlen(name) + 1;
	strcpy(t->value, value);

	return t;
}

void
_nvram_free(struct nvram_tuple *t)
{
	if (t)
		MFREE(t, sizeof(struct nvram_tuple) + strlen(t->name) + 1 + strlen(t->value) + 1);
}

int
nvram_commit(void)
{
	struct nvram_header *header;
	int ret;
	uint32 *src, *dst;
	uint i;

	if (!(header = (struct nvram_header *) MALLOC(NVRAM_SPACE))) {
		printf("nvram_commit: out of memory\n");
		return -12; /* -ENOMEM */
	}

	NVRAM_LOCK();

	/* Regenerate NVRAM */
	ret = _nvram_commit(header);
	if (ret)
		goto done;
	
	src = (uint32 *) &header[1];
	dst = src;

	for (i = sizeof(struct nvram_header); i < header->len && i < NVRAM_SPACE; i += 4)
		*dst++ = htol32(*src++);

#ifdef _CFE_
	/* Write NVRAM copy first */
	if ((ret = cfe_open(BCM_NVRAMCOPY_DEVICE)) >= 0) {
		cfe_writeblk(ret, NVRAM_SPACE_OFFSET, (unsigned char *) header, NVRAM_SPACE);
		cfe_close(ret);
	}

	/* Write main NVRAM next */
	if ((ret = cfe_open(BCM_NVRAM_DEVICE)) >= 0) {
		cfe_writeblk(ret, NVRAM_SPACE_OFFSET, (unsigned char *) header, NVRAM_SPACE);
		cfe_close(ret);
	}
#else
	if (flash_init((void *) FLASH_BASE, NULL) == 0)
		nvWrite((unsigned short *) header, NVRAM_SPACE);
#endif

 done:
	NVRAM_UNLOCK();
	MFREE(header, NVRAM_SPACE);
	return ret;
}
